# -*- coding: utf-8 -*-

"""
    This is a home for assorted utilities that didn't seem to belong elsewhere,
    stuff like the CAPO settings retriever, command line parser generator and
    so on.

"""

import argparse
from enum import Enum
import logging
import os
import pathlib
import time
from pathlib import Path
from typing import Callable, List

# pylint:disable=C0301, C0303, C0415, E0401, E0402, R0903, W0212, W1202, W0404, W0621, W1203

import psycopg2 as pg
from pycapo import CapoConfig

from .errors import (
    NoProfileException,
    MissingSettingsException,
    NGASServiceErrorException,
    SizeMismatchException,
    TERMINAL_ERROR_CODES,
)

LOG_FORMAT = "%(module)s.%(funcName)s, %(lineno)d: %(message)s"

MAX_TRIES = 10
SLEEP_INTERVAL_SECONDS = 1
FILE_SPEC_KEYS = [
    "ngas_file_id",
    "subdirectory",
    "relative_path",
    "checksum",
    "checksum_type",
    "version",
    "size",
    "server",
]
SERVER_SPEC_KEYS = ["server", "location", "cluster", "retrieve_method"]

# Prologue and epilogue for the command line parser.
_PROLOGUE = """Retrieve a product (a science product or an ancillary product) 
from the NRAO archive, either by specifying the product's locator or by 
providing the path to a product locator report."""
_EPILOGUE = TERMINAL_ERROR_CODES

# This is a dictionary of required CAPO settings
# and the attribute names we'll store them as.
REQUIRED_SETTINGS = {
    "EDU.NRAO.ARCHIVE.DATAFETCHER.DATAFETCHERSETTINGS.LOCATORSERVICEURLPREFIX": "locator_service_url",
    "EDU.NRAO.ARCHIVE.DATAFETCHER.DATAFETCHERSETTINGS.EXECUTIONSITE": "execution_site",
    "EDU.NRAO.ARCHIVE.DATAFETCHER.DATAFETCHERSETTINGS.DEFAULTTHREADSPERHOST": "threads_per_host",
    "EDU.NRAO.ARCHIVE.WORKFLOW.CONFIG.REQUESTHANDLERSETTINGS.DOWNLOADDIRECTORY": "download_dir",
}

logger = logging.getLogger(__name__)


def path_is_accessible(path: pathlib.Path) -> bool:
    """
    Is this path readable, executable, and writable?
    :param path: a file or directory
    :return:
    """

    can_access = os.access(path, os.F_OK)
    can_access = can_access and path.is_dir()
    can_access = can_access and os.access(path, os.R_OK)
    can_access = can_access and os.access(path, os.W_OK)
    can_access = can_access and os.access(path, os.X_OK)
    return can_access


def get_arg_parser():
    """Build and return an argument parser with the command line options
        for data-fetcher. this is out here and not in a class because Sphinx
        needs it
        to build the docs.

    :return: an argparse 'parser' with command line options for dataf-etcher.
    """

    cwd = pathlib.Path().absolute()
    parser = argparse.ArgumentParser(
        description=_PROLOGUE,
        epilog=_EPILOGUE,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    # Can't find a way of clearing the action groups
    # without hitting an internal attribute.
    parser._action_groups.pop()
    exclusive_group = parser.add_mutually_exclusive_group(required=True)
    exclusive_group.add_argument(
        "--product-locator",
        action="store",
        dest="product_locator",
        help="product locator to download",
    )
    exclusive_group.add_argument(
        "--location-file",
        action="store",
        dest="location_file",
        help="product locator report (in JSON)",
    )
    optional_group = parser.add_argument_group("Optional Arguments")
    optional_group.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="dry run, do not fetch product",
    )
    optional_group.add_argument(
        "--output-dir",
        action="store",
        dest="output_dir",
        default=cwd,
        help="output directory, default current dir",
    )
    optional_group.add_argument(
        "--force",
        action="store_true",
        required=False,
        dest="force",
        default=False,
        help="overwrite existing file(s) at dest",
    )
    if "CAPO_PROFILE" in os.environ:
        optional_group.add_argument(
            "--profile",
            action="store",
            dest="profile",
            help="CAPO profile to use, default '" + os.environ["CAPO_PROFILE"] + "'",
            default=os.environ["CAPO_PROFILE"],
        )
    else:
        optional_group.add_argument(
            "--profile", action="store", dest="profile", help="CAPO profile to use"
        )
    return parser


def get_capo_settings(profile: str):
    """Get the required CAPO settings for datafetcher for the provided profile
    (prod, test). Spits out an error message and exits (1) if it can't find
    one of them.

    :param profile: the profile to use
    :return: a bunch of settings
    """
    if not profile:
        raise NoProfileException("CAPO_PROFILE required; none provided")

    capo = CapoConfig(profile=profile)
    result = dict()
    for setting in REQUIRED_SETTINGS:
        setting = setting.upper()
        try:
            result[REQUIRED_SETTINGS[setting]] = capo[setting]
        except KeyError as k_err:
            raise MissingSettingsException(
                f'missing required setting "{setting}" with profile "{profile}"'
            ) from k_err
    return result


def validate_file_spec(file_spec: dict, retrieve_method_expected: bool):
    """
    Make sure this nugget of file info contains everything it should.
    :param file_spec:
    :param retrieve_method_expected:
    :return:
    """

    for key in FILE_SPEC_KEYS:
        if key not in file_spec:
            raise MissingSettingsException(f"{key} not found in file_spec")

    server = file_spec["server"]

    for key in SERVER_SPEC_KEYS:
        # if this is before retrieval mode has been set, OK not to have it
        if key not in server.keys() and retrieve_method_expected:
            raise MissingSettingsException(f"{key} not found in server spec: {server}")


class ProductLocatorLookup:

    """
    Look up the product locator for an external name (fileset ID)

    """

    def __init__(self, capo_db_settings):
        self.capo_db_settings = capo_db_settings
        self.credentials = {}
        for key, value in self.capo_db_settings.items():
            self.credentials[key.replace("metadataDatabase.", "")] = value

    def look_up_locator_for_ext_name(self, external_name) -> str:
        """
        Given a fileset ID or analogous identifier, find its product locator.

        :param external_name: e.g., ngas_fileset_id
        :return: product locator
        """

        host, dbname = self.credentials["jdbcUrl"].split(":")[2][2:].split("/")

        with pg.connect(
            dbname=dbname,
            host=host,
            user=self.credentials["jdbcUsername"],
            password=self.credentials["jdbcPassword"],
        ) as conn:
            cursor = conn.cursor()
            sql = "SELECT science_product_locator FROM science_products WHERE external_name=%s"
            cursor.execute(
                sql,
                (external_name,),
            )
            product_locator = cursor.fetchone()
        return product_locator[0]


class Retryer:
    """
    Retry executing a function, or die trying
    """

    def __init__(self, func: Callable, max_tries: int, sleep_interval: int):
        self.func = func
        self.num_tries = 0
        self.max_tries = max_tries
        self.sleep_interval = sleep_interval
        self._logger = logger
        self.complete = False

    def retry(self, *args):
        """
        Try something a specified number of times.
        Die if it doesn't work after N tries.
        :param args:
        :return:
        """
        while self.num_tries < self.max_tries and not self.complete:

            self.num_tries += 1
            exc = None
            try:
                success = self.func(args)
                if success:
                    self.complete = True
                else:
                    if self.num_tries < self.max_tries:
                        self._logger.debug(
                            "iteration #{}: {}; trying again after {} "
                            "seconds....".format(self.num_tries, exc, self.sleep_interval)
                        )
                        time.sleep(self.sleep_interval)
                    else:
                        raise NGASServiceErrorException(
                            "FAILURE after {} attempts".format(self.num_tries)
                        )

            except (NGASServiceErrorException, SizeMismatchException) as exc:
                if self.num_tries < self.max_tries:
                    self._logger.debug(
                        "{}; trying again after {} seconds....".format(exc, self.sleep_interval)
                    )
                    time.sleep(self.sleep_interval)
                else:
                    self._logger.error("FAILURE after {} attempts".format(self.num_tries))
                    raise exc


class Location(Enum):
    """
    Where the files live
    """

    DSOC = "DSOC"
    NAASC = "NAASC"

    def __str__(self):
        return str(self.value)


class Cluster(Enum):
    """
    Which cluster the files are on
    """

    DSOC = "DSOC"
    NAASC = "NAASC"

    def __str__(self):
        return str(self.value)


class ExecutionSite(Enum):
    """ Where this code is executing """

    DSOC = "DSOC"
    NAASC = "NAASC"

    def __str__(self):
        return str(self.value)


class RetrievalMode(Enum):
    """How we're retrieving a file: via streaming or via direct copy (plugin)"""

    STREAM = "stream"
    COPY = "copy"


def parse_args(args: List[str]):
    """
    Wrapper around parser.parse_args() so in the event of a foo
    an exception is thrown rather than sys.exe

    :param args:
    :return:
    """
    confirm_complete_args(args)
    to_parse = [str(arg) for arg in args]
    return get_arg_parser().parse_args(to_parse)


def confirm_complete_args(args: List[str]):
    """
    Let's scrutinize the args -before- calling parse_args()
    so we can differentiate among errors.

    :param args:
    :return:
    """

    # we must have a profile
    if "--profile" not in args and "CAPO_PROFILE" not in os.environ.keys():
        raise NoProfileException("Capo profile is required.")

    # we must have an output dir....
    if "--output-dir" not in args:
        raise MissingSettingsException("output dir is required")

    # ... and it must exist
    args_iter = iter(args)
    args_dict = dict(zip(args_iter, args_iter))

    output_dir_arg = args_dict["--output-dir"]

    output_dir = Path(output_dir_arg)
    if not output_dir.exists():
        # did they just forget to specify it?
        if output_dir_arg.startswith("--"):
            raise MissingSettingsException("--output-dir is required")
        raise FileNotFoundError(f"output dir '{output_dir_arg}' not found")
    if not path_is_accessible(output_dir):
        raise FileNotFoundError(f"permission denied on {output_dir}")

    # we must have EITHER a product locator OR a locations file....
    if "--product-locator" not in args_dict.keys() and "--location-file" not in args_dict.keys():
        raise MissingSettingsException("either product locator or locations file required")
    if "--product-locator" in args_dict.keys() and "--location-file" in args_dict.keys():
        raise MissingSettingsException("product locator OR locations file required -- not both")

    if "--product-locator" in args:
        product_locator = args_dict["--product-locator"]
        assert product_locator
    else:
        locations_file = args_dict["--location-file"]
        if not Path(locations_file).exists():
            raise FileNotFoundError(f"{locations_file} not found")
