from __future__ import annotations

import datetime
import pathlib
from typing import Tuple

import pendulum
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from workspaces.capability.enums import CapabilityRequestState, CapabilityStepType
from workspaces.capability.helpers import CapabilitySequence, CapabilityStep
from workspaces.capability.helpers_interfaces import CapabilitySequenceIF, ParameterIF
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityIF,
    CapabilityRequestIF,
    CapabilityVersionIF,
)
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import JSONSerializable

Base = declarative_base()


class Capability(Base, CapabilityIF):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """

    __tablename__ = "capabilities"
    name = sa.Column("capability_name", sa.String, primary_key=True)
    steps = sa.Column("capability_steps", sa.String)
    max_jobs = sa.Column("max_jobs", sa.Integer)
    enabled = sa.Column("enabled", sa.Boolean, default=True, server_default="true")
    requests = relationship("CapabilityRequest", back_populates="capability")

    @classmethod
    def from_file(cls, filename: str) -> Capability:
        """
        Method that creates an instance of a Capability from a given file

        :param filename:
        :return: Capability instance
        """
        self = cls()
        self.name, self.max_jobs, self.steps = cls.parse_capability_file(filename)
        return self

    @staticmethod
    def parse_capability_file(
        filename: str,
    ) -> Tuple[str, int, CapabilitySequenceIF]:
        """
        Parse a file containing metadata for a capability

        :param filename:
        :return: Tuple containing capability's ID, name, max concurrent jobs, and sequence of steps
        """
        with open(filename, "r") as f:
            name, max_jobs = f.readline().split(" ")
            steps = []

            for line in f.readlines():
                step_type, step_value = line.strip().split(" ")
                steps.append(CapabilityStep(CapabilityStepType.from_string(step_type), step_value))

            return name, int(max_jobs), CapabilitySequence(steps)

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF):
        """
        Create a new request for this capability with specific options

        FIXME: Future products needs a string representation
        """
        request = CapabilityRequest(capability=self, parameters=str(parameters))
        self.requests.append(request)
        return request

    def __str__(self) -> str:
        return (
            f"Capability: {self.name}, max concurrent jobs of {self.max_jobs}"
            f"\nSequence: {self.steps}"
        )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "name": self.name,
            "max_jobs": self.max_jobs,
            "steps": str(self.steps),
            "enabled": self.enabled,
        }

    @classmethod
    def from_json(cls, json: dict) -> Capability:
        """
        Method that creates an instance of a Capability from given JSON

        :param json: the description of this capabiity in a dictionary
        :return: Capability instance
        """
        copy = dict(json)
        del copy["type"]
        return Capability(**copy)


class CapabilityRequest(Base, CapabilityRequestIF):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """

    __tablename__ = "capability_requests"
    id = sa.Column("capability_request_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_name = sa.Column(
        "capability_name", sa.String, sa.ForeignKey("capabilities.capability_name")
    )
    parameters = sa.Column("parameters", sa.JSON)
    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )

    # FIXME: This needs to be changed to properly keep track of product locators.
    # future_products = sa.Column('future_products', sa.String)
    versions = relationship("CapabilityVersion", back_populates="request")
    capability = relationship(Capability, back_populates="requests")

    def update_state(self, state: CapabilityRequestState):
        self.state = state.name

    def __str__(self):
        return f"CapabilityRequest object: {self.__dict__}"

    @property
    def created_at_ago(self) -> str:
        return pendulum.instance(self.created_at).diff_for_humans()

    @property
    def updated_at_ago(self) -> str:
        return pendulum.instance(self.updated_at).diff_for_humans()

    def __repr__(self):
        return (
            f"<CapabilityRequest #{self.id} "
            f"for capability {self.capability_name}, "
            f"created {self.created_at_ago} and "
            f"last updated {self.updated_at_ago})>"
        )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "id": self.id,
            "capability_name": self.capability_name,
            "state": self.state,
            "parameters": self.parameters,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
        }

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityRequest(**copy)


class CapabilityVersion(Base, CapabilityVersionIF):
    """
    Schema representation of a capability version
    """

    __tablename__ = "capability_versions"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
        primary_key=True,
    )
    version_number = sa.Column("version_number", sa.Integer, primary_key=True)
    parameters = sa.Column("parameters", sa.JSON)
    request = relationship(CapabilityRequest, back_populates="versions")
    executions = relationship("CapabilityExecution", back_populates="version")

    @property
    def capability(self):
        return self.request.capability


class SaPath(sa.types.TypeDecorator):
    impl = sa.types.String

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return pathlib.Path(value)

    def copy(self, **kw):
        return SaPath(self.impl.length)


class CapabilityExecution(Base, CapabilityExecutionIF, JSONSerializable):
    """
    Schema representation of a capability request's execution record
    """

    __tablename__ = "capability_executions"

    id = sa.Column("execution_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
    )
    version_number = sa.Column("capability_version_number", sa.Integer)
    current_step = sa.Column("current_step", sa.Integer)
    steps = sa.Column("steps", sa.String)
    version = relationship(CapabilityVersion, back_populates="executions")
    current_workflow_request_id = sa.Column("current_workflow_request_id", sa.Integer)
    delivery_url = sa.Column("delivery_url", sa.String)
    delivery_path = sa.Column("delivery_path", SaPath)

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        server_onupdate=sa.func.now(),
        nullable=False,
    )

    __table_args__ = (
        sa.ForeignKeyConstraint(
            (capability_request_id, version_number),
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
    )

    def update_execution(self, workflow_request_id: int):
        self.current_workflow_request_id = workflow_request_id

    @property
    def capability(self):
        return self.version.capability

    @property
    def capability_request(self):
        return self.version.request

    def on_last_step(self) -> bool:
        return self.current_step == len(CapabilitySequence.from_str(self.steps)) - 1

    def __lt__(self, other):
        return self.created_at < other.created_at

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return dict(
            type=self.__class__.__name__,
            id=self.id,
            state=self.state,
            capability_request_id=self.capability_request_id,
            version_number=self.version_number,
            current_step=self.current_step,
            current_workflow_request_id=self.current_workflow_request_id,
            created_at=self.created_at.isoformat(),
            updated_at=self.updated_at.isoformat(),
        )

    @classmethod
    def from_json(cls, json: dict) -> any:
        copy = dict(json)
        del copy["type"]
        copy["created_at"] = datetime.datetime.fromisoformat(copy["created_at"])
        copy["updated_at"] = datetime.datetime.fromisoformat(copy["updated_at"])
        return CapabilityExecution(**copy)
